/*
 * *********************************************************************************************************************
 *
 *          !!
 *        .F88X
 *        X8888Y
 *      .}888888N;
 *        i888888N;        .:!              .I$WI:
 *          R888888I      .'N88~            i8}+8Y&8"l8i$8>8W~'>W8}8]KW+8IIN"8&
 *          .R888888I    .;N8888~          .X8'  "8I.!,/8"  !%NY8`"8I8~~8>,88I
 *            +888888N;  .8888888Y                                  "&&8Y.}8,
 *            ./888888N;  .R888888Y        .'}~    .>}'.`+>  i}!    "i'  +/'  .'i~  !11,.:">,  .~]!  .i}i
 *              ~888888%:  .I888888l      .]88~`1/iY88Ii+1'.R$8$8]"888888888>  Y8$  W8E  X8E  W8888'188Il}Y88$*
 *              18888888    E8888881    .]W%8$`R8X'&8%++N8i,8N%N8+l8%`  .}8N:.R$RE%N88N%N$K$R  188,FE$8%~Y88I
 *            .E888888I  .i8888888'      .:$8I;88+`E8R:/8N,.>881.`$8E/1/]N8X.Y8N`"KF&&FK!'88*."88K./$88%RN888+~
 *            8888888I  .,N888888~        ~88i"8W,!N8*.I88.}888%F,i$88"F88"  888:E8X.>88!i88>`888*.}Fl1]*}1YKi'
 *          i888888N'      I888Y          ]88;/EX*IFKFK88X  K8R  .l8W  88Y  ~88}'88E&%8W.X8N``]88!.$8K  .:W8I
 *        .i888888N;        I8Y          .&8$  .X88!  i881.:%888>I88  ;88]  +88+.';;;;:.Y88X  18N.,88l  .+88/
 *      .:R888888I
 *      .&888888I                                          Copyright (c) 2009-2020.  博锐尚格科技股份有限公司
 *        ~8888'
 *        .!88~                                                                     All rights reserved.
 *
 * *********************************************************************************************************************
 */

package com.persagy.mybatis

import com.persagy.base.extensions.getAllDeclaredFields
import com.persagy.base.extensions.toJson
import com.persagy.database.SAbstractDao
import com.persagy.database.SAbstractFunction
import com.persagy.database.SFilter
import com.persagy.database.SQueryBuilder
import com.persagy.database.enums.SOps
import com.persagy.gis.entity.SAbstractGisGeometry
import org.apache.ibatis.jdbc.SQL
import org.slf4j.LoggerFactory
import javax.persistence.Column

/**
 * 数据库驱动
 *
 * @author  庞利祥 <sybotan@126.com>
 */
abstract class SAbstractDriver {
    /** 静态对象 */
    companion object {
        /** 日志记录器 */
        private val logger = LoggerFactory.getLogger(SAbstractDriver::class.java)
    }

    /**
     * 在数据库中插入实体
     *
     * @param argsMap         注入参数
     * @return 插入 SQL 语句
     */
    abstract fun insert(argsMap: Map<String, Any?>): String

    /**
     * 在数据库中替换实体
     *
     * @param argsMap         注入参数
     *
     * @return 替换 SQL 语句
     */
    abstract fun replace(argsMap: Map<String, Any?>): String

    /**
     * 根据条件删除
     *
     * @param argsMap         注入参数
     * @return 删除 SQL 语句
     */
    abstract fun delete(argsMap: Map<String, Any?>): String

    /**
     * 根据条件更新
     *
     * @param argsMap         注入参数
     * @return 更新 SQL 语句
     */
    abstract fun update(argsMap: Map<String, Any?>): String

    /**
     * 删除所有记录
     *
     * @param argsMap         注入参数
     * @return 删除 SQL 语句
     */
    abstract fun deleteAll(argsMap: Map<String, Any?>): String

    /**
     * 执行查询操作
     *
     * @param argsMap         注入参数
     * @return 查询 SQL 语句
     */
    @Suppress("UNCHECKED_CAST")
    abstract fun execQuery(argsMap: Map<String, Any?>): String


    /**
     * 执行查询操作
     *
     * @param argsMap         注入参数
     * @return 查询 SQL 语句
     */
    @Suppress("UNCHECKED_CAST")
    fun execCount(argsMap: Map<String, Any?>): String {
        val entityClass = argsMap["entityClass"] as Class<*>
        val builder = argsMap["builder"] as SQueryBuilder<*>
        val whereArgs = argsMap["whereArgs"] as HashMap<String, Any>

        val clause = buildWhereArgList(entityClass, builder.filterList, whereArgs)

        val sql = object: SQL() {
            init {
                SELECT("count(1) AS total")

                /** 如果查询构建器指定了表名 */
                if (builder.tableName.isNullOrEmpty()) {
                    FROM(SSqlProvider.driver.tableName(entityClass))
                } else {
                    FROM(builder.tableName!!)
                }

                /** 过滤条件 */
                if (clause.isNotEmpty()) {
                    WHERE(clause)
                }

                /** 分组 */
                if (!builder.group.nameList.isNullOrEmpty()) {
                    for (name in builder.group.nameList) {
                        val colName = SAbstractDao.columnName(entityClass, name)
                        GROUP_BY(SSqlProvider.driver.escName(colName))
                    }
                }
            }
        }.toString()
        logger.debug("SQL= $sql")
        return sql
    }

    /**
     * 获得表名，取实体类 @Table 注解的名称，如果没有注解，则取实体类类名。
     *
     * @param clz         mapper 类类型
     * @return 表名
     */
    abstract fun tableName(clz: Class<*>): String

    /**
     * 将 name 加上引用符号
     *
     * @param name        名称
     * @return 加上引用符号的名称
     */
    abstract fun escName(name: String): String

    /**
     * 实体类的值，如果不是基本类型，则转换为 json
     *
     * @param entity      实体类
     * @param name        名称
     * @return 加上引用符号的名称
     */
    abstract fun entityValue(entity: HashMap<String, Any?>, name: String): String

    /**
     * 生成 limit 语句
     *
     * @param offset      偏移位置
     * @param count       取得的记录数
     * @return LIMIT 子句
     */
    abstract fun limit(offset: Int?, count: Int?): String

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    //
    /**
     * 转换 where 字句参数
     *
     * @param argName     参数名
     * @param argValue    参数值
     * @param ops         运算符
     * @return 转换后的 SQL
     */
    abstract fun whereArg(argName: String, argValue: String, ops: SOps): String


    /**
     * 转换参数名 SQL 语句
     *
     * @param name        字段名
     * @param path        json 路径，如果不是 json 对象，则该值为空
     * @return 转换后的 SQL
     */
    abstract fun argName(name: String, path: String?): String

    /**
     * 转换参数值 SQL 语句
     *
     * @param argName     参数名
     * @param isJson      是否 Json 对象
     * @return 转换后的 SQL
     */
    abstract fun argValue(argName: String, isJson: Boolean): String

    /**
     * 转换 IN 条件表达式
     *
     * @param name        字段名
     * @param path        json 路径，如果字段不为 json，则 path 为 null
     * @param inArg       in 操作参数
     * @return 转换后的 SQL
     */
    abstract fun inArg(name: String, path: String?, inArg: String): String

    /**
     * 将名称及 json 路径转换为 SQL
     *
     * @param name        字段名
     * @param path        json 路径，如果字段不为 json，则 path 为 null
     * @return 转换后的 SQL
     */
    abstract fun columnName(name: String, path: String?): String

    /**
     * 将参数值转换为 SQL 语句
     *
     * @param value       值
     * @param isJson      值是否转换为 json 格式
     * @return 转换后的 SQL
     */
    abstract fun argValueToJson(value: String, isJson: Boolean): String

    /**
     * 更新 Json 类型字段
     *
     * @param name        字段名
     * @param value       值
     * @return 更新 SQL 字符串
     */
    abstract fun updateJson(name: String, value: Any): String

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    // 函数
    /**
     * 处理函数支持
     *
     * @param func        函数对象
     * @param colName     字段名
     * @return 生成的 SQL 语句
     */
    abstract fun processFunction(func: SAbstractFunction, colName: String): String;

    /**
     * jsonb 中属性置空
     *
     * @param nullItem  表列名
     */
    abstract fun updateNullItem(nullItem: String): String

    /**
     * 构造 where 子句参数
     *
     * @param filterList      过滤条件
     * @param values          子句参数列表
     * @return 返回 where 子句
     */
    protected fun buildWhereArgList(clazz: Class<*>, filterList: List<SFilter>, values: HashMap<String, Any>): String {
        val builder = StringBuilder()

        for (filter in filterList) {
            val where = buildWhereArg(clazz, filter, values)
            if (builder.isNotEmpty()) {
                builder.append(" AND ")
            }
            builder.append(where)
        }

        val sql = builder.toString()
//        logger.debug("buildWhereArgList = $sql")
        return sql
    }

    /**
     * 构造 where 子句参数
     *
     * @param clazz           对应的实体类类型
     * @param filter          过滤条件
     * @param values          子句参数列表
     * @return 返回 where 子句
     */
    protected fun buildWhereArg(clazz: Class<*>, filter: SFilter, values: HashMap<String, Any>): String {
        val builder = StringBuilder()
        val argName = "arg${values.size}"
        when (filter.opt) {
            SOps.EQ,            /** 等于 */
            SOps.NE,            /** 不等于 */
            SOps.GT,            /** 大于 */
            SOps.GTE,           /** 大于等于 */
            SOps.LT,            /** 小于 */
            SOps.LTE,           /** 小于等于 */
            SOps.CONTAIN,       /** 包含 */
            SOps.STARTWITH,     /** 以指定字符串开始 */
            SOps.ENDWITH,       /** 以指定字符串结束 */
            SOps.ISNULL         /** 为空 */
            -> {
                var path: String? = null
                var propertyName = if (filter.first is SAbstractFunction) {
                    (filter.first!! as SAbstractFunction).name
                } else {
                    filter.first as String
                }
                val pos = propertyName.indexOf(".")
                if (pos > 0) {
                    path = propertyName.substring(pos + 1)
                    propertyName = propertyName.substring(0, pos)
                }
                val colName = SAbstractDao.columnName(clazz, propertyName)
                try {
                    values[argName] = filter.second!!
                } catch (e: Exception) {
                    // DO NOTHING
                }

                val isJson = if (filter.opt == SOps.CONTAIN || filter.opt == SOps.STARTWITH || filter.opt == SOps.ENDWITH) {
                    false
                } else {
                    path != null
                }

                return if (filter.first is SAbstractFunction) {
                    SSqlProvider.driver.whereArg(SSqlProvider.driver.processFunction(filter.first!! as SAbstractFunction, SSqlProvider.driver.argName(colName, path)), SSqlProvider.driver.argValue(argName, isJson), filter.opt)
                } else {
                    SSqlProvider.driver.whereArg(SSqlProvider.driver.argName(colName, path), SSqlProvider.driver.argValue(argName, isJson), filter.opt)
                }
            }
            SOps.IN -> {         /** 在列表内 */
                /** 如果列表不为空，则添加IN子句 */
                if (filter.second !=null && List::class.java.isAssignableFrom(filter.second!!.javaClass) &&
                        (filter.second as List<Any?>).size > 0) {
                    var path: String? = null
                    val inArg = buildInArgs(filter, values, argName)

                    if (filter.first is SAbstractFunction) {
                        val colName = SAbstractDao.columnName(clazz, (filter.first as SAbstractFunction).name)
                        val col = SSqlProvider.driver.processFunction(filter.first!! as SAbstractFunction, SSqlProvider.driver.argName(colName, path))
                        builder.append("$col IN ($inArg)")
                    } else {
                        var propertyName = filter.first!!.toString()
                        val pos = propertyName.indexOf(".")
                        if (pos > 0) {
                            path = propertyName.substring(pos + 1)
                            propertyName = propertyName.substring(0, pos)
                        }
                        val colName = SAbstractDao.columnName(clazz, propertyName)
                        builder.append(SSqlProvider.driver.inArg(colName, path, inArg))
                    }
                    //"${driver.keyEsc}$colName${driver.keyEsc} IN ($inArg)"
                }
            }
            SOps.NOT -> {        /** 逻辑非 */
                builder.append("NOT (${buildWhereArg(clazz, filter.first as SFilter, values)})")
            }
            SOps.AND -> {        /** 逻辑与 */
                builder.append("(${buildWhereArg(clazz, filter.first as SFilter, values)}) AND (${buildWhereArg(clazz, filter.second as SFilter, values)})")
            }
            SOps.OR -> {         /** 逻辑或 */
                builder.append("((${buildWhereArg(clazz, filter.first as SFilter, values)}) OR (${buildWhereArg(clazz, filter.second as SFilter, values)}))")
            }
            else -> {
            }
        }

        return builder.toString()
    }

    /**
     * 生成 IN 操作参数
     *
     * @param filter        条件
     * @param values        值列表
     * @param argPrefix     参数前缀
     */
    @Suppress("UNCHECKED_CAST")
    protected fun buildInArgs(filter: SFilter, values: HashMap<String, Any>, argPrefix: String): String {
        val builder = StringBuilder()
        val argList = filter.second!! as List<Any>
        val property = filter.first?.toString()
        val isJson = if (filter.first is SAbstractFunction) {
            false
        } else {
            property?.indexOf(".") != -1
        }

        for ((index, item) in argList.withIndex()) {
            if (index > 0) {
                builder.append(", ")
            }
            builder.append(SSqlProvider.driver.argValueToJson("#{whereArgs.${argPrefix}_$index}", isJson))
            values["${argPrefix}_$index"] = item
        }

        return builder.toString()
    }

    /**
     * 构建 columns 值，如果不指定字段列表，则查所有记录
     *
     * @param clazz       实体类类型
     * @param fieldList   属性映射列表
     * @return values 列表
     */
    protected fun buildColumns(clazz: Class<*>, fieldList: ArrayList<String>?): String {
        /** 如果不指定字段列表，则查所有记录 */
//        if (fieldList == null || fieldList.size <= 0) {
//            return "*"
//        }
        logger.debug("222222222222222222222222222222222222")
        val builder = StringBuilder()
        for (field in clazz.getAllDeclaredFields()) {
            logger.debug("field = ${field.toJson()}")
            if (fieldList != null && fieldList.size > 0 && !fieldList.contains(field.name)) {
                continue
            }
            val anna = field.getAnnotation(Column::class.java) ?: continue

            if (builder.isNotEmpty()) {
                builder.append(", ")
            }
            if (SAbstractGisGeometry::class.java.isAssignableFrom(field.type)) {
                val col = SSqlProvider.driver.escName(anna.name)
                builder.append("ST_AsText($col) AS $col")
            }else {
                builder.append(SSqlProvider.driver.escName(anna.name))
            }

        }
        return builder.toString()
    }
}
